# A server that returns a message that cannot be parsed as JSON.
import json

from lsprotocol import types
from pygls.lsp.server import LanguageServer

server = LanguageServer(name="completion-exit-server", version="v1.0")


def bad_send_data(data):
    """Sends data to the client in a way that cannot be parsed."""
    if not data:
        return

    self = server.protocol
    body = json.dumps(data, default=self._serialize_message)
    body = body.replace('"', "'").encode(self.CHARSET)
    header = (
        f"Content-Length: {len(body)}\r\n"
        f"Content-Type: {self.CONTENT_TYPE}; charset={self.CHARSET}\r\n\r\n"
    ).encode(self.CHARSET)

    self.transport.write(header + body)


@server.feature(types.TEXT_DOCUMENT_COMPLETION)
def on_complete(server: LanguageServer, params: types.CompletionParams):
    server.protocol._send_data = bad_send_data
    return [types.CompletionItem(label="item-one")]


if __name__ == "__main__":
    server.start_io()



from __future__ import annotations

import asyncio
import logging
import typing
from concurrent.futures import Future

from pygls.protocol import LanguageServerProtocol

from .checks import check_params_against_client_capabilities
from .checks import check_result_against_client_capabilities

if typing.TYPE_CHECKING:
    from .client import LanguageClient


logger = logging.getLogger(__name__)


class LanguageClientProtocol(LanguageServerProtocol):
    """An extended protocol class adding functionality useful for testing."""

    _server: LanguageClient  # type: ignore[assignment]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._notification_futures = {}

    def _handle_request(self, msg_id, method_name, params):
        """Wrap pygls' handle_request implementation. This will

        - Check if the request from the server is compatible with the client's stated
          capabilities.

        """
        check_params_against_client_capabilities(
            self._server.capabilities, method_name, params
        )
        return super()._handle_request(msg_id, method_name, params)

    def _handle_notification(self, method_name, params):
        """Wrap pygls' handle_notification implementation. This will

        - Notify a future waiting on the notification, if applicable.

        - Check the params to see if they are compatible with the client's stated
          capabilities.

        """
        future = self._notification_futures.pop(method_name, None)
        if future:
            future.set_result(params)

        super()._handle_notification(method_name, params)

    async def send_request_async(self, method, params=None):
        """Wrap pygls' ``send_request_async`` implementation. This will

        - Check the params to see if they're compatible with the client's stated
          capabilities
        - Check the result to see if it's compatible with the client's stated
          capabilities

        Parameters
        ----------
        method
           The method name of the request to send

        params
           The associated parameters to go with the request

        Returns
        -------
        Any
           The result
        """
        check_params_against_client_capabilities(
            self._server.capabilities, method, params
        )
        result = await super().send_request_async(method, params)
        check_result_against_client_capabilities(
            self._server.capabilities,
            method,
            result,  # type: ignore
        )

        return result

    def wait_for_notification(self, method: str, callback=None) -> Future:
        """Wait for a notification message with the given ``method``.

        Parameters
        ----------
        method
           The method name to wait for

        callback
           If given, ``callback`` will be called with the notification message's
           ``params`` when recevied

        Returns
        -------
        Future
           A future that will resolve when the requested notification message is
          recevied.
        """
        future: Future = Future()
        if callback:

            def wrapper(future: Future):
                result = future.result()
                callback(result)

            future.add_done_callback(wrapper)

        self._notification_futures[method] = future
        return future

    def wait_for_notification_async(self, method: str):
        """Wait for a notification message with the given ``method``.

        Parameters
        ----------
        method
           The method name to wait for

        Returns
        -------
        Any
           The notification message's ``params``
        """
        future = self.wait_for_notification(method)
        return asyncio.wrap_future(future)



from lsprotocol import types
from pygls.lsp.server import LanguageServer

server = LanguageServer("window-show-document", "v1")


@server.feature(types.TEXT_DOCUMENT_COMPLETION)
async def completion(ls: LanguageServer, params: types.CompletionParams):
    items = []
    await ls.window_show_document_async(
        types.ShowDocumentParams(uri=params.text_document.uri)
    )

    for i in range(10):
        items.append(types.CompletionItem(label=f"item-{i}"))

    return items


if __name__ == "__main__":
    server.start_io()



from unittest.mock import Mock

from lsprotocol import types
from pygls.lsp.server import LanguageServer

server = LanguageServer("window-create-progress", "v1")


@server.command("do.progress")
async def do_progress(ls: LanguageServer, *args):
    token = "a-token"

    await ls.work_done_progress.create_async(token)

    # Begin
    ls.work_done_progress.begin(
        token,
        types.WorkDoneProgressBegin(title="Indexing", percentage=0),
    )
    # Report
    for i in range(1, 4):
        ls.work_done_progress.report(
            token,
            types.WorkDoneProgressReport(message=f"{i * 25}%", percentage=i * 25),
        )
    # End
    ls.work_done_progress.end(token, types.WorkDoneProgressEnd(message="Finished"))

    return "a result"


@server.command("duplicate.progress")
async def duplicate_progress(ls: LanguageServer, *args):
    token = "duplicate-token"

    # Need to stop pygls preventing us from using the progress API wrong.
    ls.work_done_progress._check_token_registered = Mock()
    await ls.work_done_progress.create_async(token)

    # pytest-lsp should return an error here.
    await ls.work_done_progress.create_async(token)


@server.command("no.progress")
async def no_progress(ls: LanguageServer, *args):
    token = "undefined-token"

    # Begin
    ls.work_done_progress.begin(
        token,
        types.WorkDoneProgressBegin(title="Indexing", percentage=0, cancellable=False),
    )
    # Report
    for i in range(1, 4):
        ls.work_done_progress.report(
            token,
            types.WorkDoneProgressReport(message=f"{i * 25}%", percentage=i * 25),
        )
    # End
    ls.work_done_progress.end(token, types.WorkDoneProgressEnd(message="Finished"))


if __name__ == "__main__":
    server.start_io()



import sys

from lsprotocol import types
from pygls.lsp.server import LanguageServer

server = LanguageServer("server-stderr", "v1")


@server.feature(types.TEXT_DOCUMENT_COMPLETION)
def completion(params: types.CompletionParams):
    items = []

    for i in range(10):
        print(f"Suggesting item {i}", file=sys.stderr, flush=True)
        items.append(types.CompletionItem(label=f"item-{i}"))

    return items


if __name__ == "__main__":
    server.start_io()



import pathlib
import sys
from typing import Any
from typing import Dict
from typing import List

import pygls.uris as uri
import pytest

from pytest_lsp.plugin import ClientServerConfig


@pytest.mark.parametrize(
    "config, kwargs, expected",
    [
        (ClientServerConfig(server_command=["command"]), {}, ["command"]),
        (
            ClientServerConfig(server_command=["command"]),
            {"devtools": "1234"},
            [
                "lsp-devtools",
                "agent",
                "--host",
                "localhost",
                "--port",
                "1234",
                "--",
                "command",
            ],
        ),
        (
            ClientServerConfig(server_command=["command"]),
            {"devtools": "localhost:1234"},
            [
                "lsp-devtools",
                "agent",
                "--host",
                "localhost",
                "--port",
                "1234",
                "--",
                "command",
            ],
        ),
        (
            ClientServerConfig(server_command=["command"]),
            {"devtools": "127.0.0.1:1234"},
            [
                "lsp-devtools",
                "agent",
                "--host",
                "127.0.0.1",
                "--port",
                "1234",
                "--",
                "command",
            ],
        ),
    ],
)
def test_get_server_command(
    config: ClientServerConfig, kwargs: Dict[str, Any], expected: List[str]
):
    """Ensure that we can build the server start command correctly."""
    actual = config.get_server_command(**kwargs)
    assert expected == actual


def setup_test(pytester: pytest.Pytester, server_name: str, test_code: str):
    """Boilerplate for setting up a test."""

    python = sys.executable
    testdir = pathlib.Path(__file__).parent

    server = testdir / "servers" / server_name
    root_uri = uri.from_fs_path(str(testdir))

    pytester.makeini(
        """\
[pytest]
asyncio_mode = auto
asyncio_default_fixture_loop_scope = function
"""
    )

    pytester.makeconftest(
        f"""
from lsprotocol.types import InitializeParams

import pytest_lsp
from pytest_lsp import ClientServerConfig
from pytest_lsp import LanguageClient
from pytest_lsp import client_capabilities


@pytest_lsp.fixture(
    config=ClientServerConfig(
        server_command=[r"{python}", r"{server}"],
    )
)
async def client(lsp_client: LanguageClient):
    await lsp_client.initialize_session(
        InitializeParams(
            capabilities=client_capabilities("visual-studio-code"),
            root_uri=r"{root_uri}"
        )
    )
    yield

    await lsp_client.shutdown_session()
    """
    )

    pytester.makepyfile(test_code)


def test_detect_server_exit(pytester: pytest.Pytester):
    """Ensure that the plugin can detect when the server process exits."""

    test_code = """\
import pytest

@pytest.mark.asyncio
async def test_capabilities(client):
    ...  # Test code does not matter in this case, as the error should be in the setup.
"""

    setup_test(pytester, "hello.py", test_code)
    results = pytester.runpytest("-vv")

    results.assert_outcomes(errors=1)

    message = r"E\s+RuntimeError: Server process \d+ exited with code: 0"
    results.stdout.re_match_lines(message)


def test_detect_server_exit_mid_request(pytester: pytest.Pytester):
    """Ensure that the plugin can detect when the server process exits mid request."""

    test_code = """\
import pytest
from lsprotocol.types import CompletionParams
from lsprotocol.types import Position
from lsprotocol.types import TextDocumentIdentifier


@pytest.mark.asyncio
async def test_capabilities(client):
    expected = {str(i) for i in range(10)}

    for i in range(10):
        params = CompletionParams(
            text_document=TextDocumentIdentifier(uri="file:///test.txt"),
            position=Position(line=0, character=0)
        )
        items = await client.text_document_completion_async(params)
        assert len({i.label for i in items} & expected) == len(items)
"""

    setup_test(pytester, "completion_exit.py", test_code)
    results = pytester.runpytest("-vv")

    results.assert_outcomes(failed=1, errors=1)

    message = r"E\s+RuntimeError: Server process \d+ exited with code: 0"
    results.stdout.re_match_lines(message)
    results.stdout.fnmatch_lines("E*RuntimeError: Client has been stopped.")


def test_detect_server_crash(pytester: pytest.Pytester):
    """Ensure the plugin can detect when the server process crashes on boot."""

    test_code = """\
import pytest

@pytest.mark.asyncio
async def test_capabilities(client):
    ... # Test code does not matter in this case, as the error should be in the setup.
"""

    setup_test(pytester, "crash.py", test_code)
    results = pytester.runpytest("-vv")

    results.assert_outcomes(errors=1)

    message = r"E\s+RuntimeError: Server process \d+ exited with code: 1"
    results.stdout.re_match_lines(message)
    results.stdout.fnmatch_lines("ZeroDivisionError: division by zero")


def test_detect_invalid_json(pytester: pytest.Pytester):
    """Ensure that the plugin can detect when the server sends bad JSON."""

    test_code = """\
import pytest
from lsprotocol.types import CompletionParams
from lsprotocol.types import Position
from lsprotocol.types import TextDocumentIdentifier


@pytest.mark.asyncio
async def test_capabilities(client):
    expected = {str(i) for i in range(10)}

    for i in range(10):
        items = await client.text_document_completion_async(
            CompletionParams(
                text_document=TextDocumentIdentifier(uri="file:///test.txt"),
                position=Position(line=0, character=0)
            )
        )
        assert len({i.label for i in items} & expected) == len(items)
"""

    setup_test(pytester, "invalid_json.py", test_code)
    results = pytester.runpytest("-vv")

    results.assert_outcomes(errors=1, failed=1)

    if sys.version_info < (3, 9):
        message = "E*CancelledError"
    else:
        message = "E*asyncio.exceptions.CancelledError: JsonRpcInternalError: *"

    results.stdout.fnmatch_lines(message)



from lsprotocol import types
from pygls.lsp.server import LanguageServer

server = LanguageServer("window-log-message", "v1")


@server.feature(types.TEXT_DOCUMENT_COMPLETION)
def completion(ls: LanguageServer, params: types.CompletionParams):
    items = []

    for i in range(10):
        ls.window_log_message(
            types.LogMessageParams(
                message=f"Suggesting item {i}",
                type=types.MessageType.Log,
            )
        )
        items.append(types.CompletionItem(label=f"item-{i}"))

    return items


if __name__ == "__main__":
    server.start_io()



import sys

import pytest
import pytest_lsp
from lsprotocol import types
from pytest_lsp import ClientServerConfig, LanguageClient


@pytest_lsp.fixture(
    scope="module",
    config=ClientServerConfig(server_command=[sys.executable, "server.py"]),
)
async def client(lsp_client: LanguageClient):
    # Setup
    params = types.InitializeParams(capabilities=types.ClientCapabilities())
    await lsp_client.initialize_session(params)

    yield

    # Teardown
    await lsp_client.shutdown_session()


@pytest.mark.asyncio(loop_scope="module")
async def test_completion_hello(client: LanguageClient):
    """Ensure that the server implements completions correctly."""

    results = await client.text_document_completion_async(
        params=types.CompletionParams(
            position=types.Position(line=1, character=0),
            text_document=types.TextDocumentIdentifier(uri="file:///path/to/file.txt"),
        )
    )
    assert results is not None

    if isinstance(results, types.CompletionList):
        items = results.items
    else:
        items = results

    labels = {item.label for item in items}
    assert "hello" in labels


@pytest.mark.asyncio(loop_scope="module")
async def test_completion_world(client: LanguageClient):
    """Ensure that the server implements completions correctly."""

    results = await client.text_document_completion_async(
        params=types.CompletionParams(
            position=types.Position(line=1, character=0),
            text_document=types.TextDocumentIdentifier(uri="file:///path/to/file.txt"),
        )
    )
    assert results is not None

    if isinstance(results, types.CompletionList):
        items = results.items
    else:
        items = results

    labels = {item.label for item in items}
    assert "world" in labels



from __future__ import annotations

import inspect
import logging
import sys
import textwrap
import typing

import attrs
import pytest
import pytest_asyncio
from pygls.client import JsonRPCClient

from pytest_lsp.client import LanguageClient
from pytest_lsp.client import make_test_lsp_client

if typing.TYPE_CHECKING:
    from typing import Any
    from typing import Callable


logger = logging.getLogger("client")


@attrs.define
class ClientServerConfig:
    """Configuration for a Client-Server connection."""

    server_command: list[str]
    """The command to use to start the language server."""

    client_factory: Callable[[], JsonRPCClient] = attrs.field(
        default=make_test_lsp_client,
    )
    """Factory function to use when constructing the test client instance."""

    server_env: dict[str, str] | None = attrs.field(default=None)
    """Environment variables to set when starting the server."""

    def _get_devtools_command(self, server: str) -> list[str]:
        """Get the lsp-devtools command required to connect to the given ``server``"""

        if ":" in server:
            host, port = server.split(":")
        else:
            host, port = "localhost", server

        try:
            int(port)
        except ValueError as e:
            raise ValueError(f"Invalid port number: {port!r}") from e

        return ["lsp-devtools", "agent", "--host", host, "--port", port, "--"]

    def get_server_command(self, devtools: str | None = None) -> list[str]:
        """Get the command to start the server with."""
        server_command = []
        if devtools is not None:
            server_command.extend(self._get_devtools_command(devtools))

        server_command.extend(self.server_command)

        return server_command

    async def start(self, devtools: str | None = None) -> JsonRPCClient:
        """Return the client instance to use for the test.

        Parameters
        ----------
        devtools
           If set, enable ``lsp-devtools`` integration, should be of the form
           ``<port>`` or ``<host>:<port>``. Where ``<host>`` and ``<port>``
           describe how to connect to an ``lsp-devtools`` server program.

        Returns
        -------
        JsonRPCClient
           The client instance to use in the test.
        """
        client = self.client_factory()
        server_command = self.get_server_command(devtools)

        await client.start_io(*server_command, env=self.server_env)
        return client


def pytest_addoption(parser):
    group = parser.getgroup("lsp")
    group.addoption(
        "--lsp-devtools",
        dest="devtools",
        nargs="?",
        action="store",
        default=None,
        const="localhost:8765",
        help="Enable lsp-devtools integration.",
    )


def pytest_runtest_makereport(item: pytest.Item, call: pytest.CallInfo):
    """Add any captured log messages to the report."""
    client: LanguageClient | None = None

    if not hasattr(item, "funcargs"):
        return

    for arg in item.funcargs.values():
        if isinstance(arg, LanguageClient):
            client = arg
            break

    if not client:
        return

    levels = ["ERROR: ", " WARN: ", " INFO: ", "  LOG: "]

    if call.when == "setup":
        captured_messages = client.log_messages[: client._setup_log_index + 1]
    else:
        captured_messages = client.log_messages[client._last_log_index :]
        client._last_log_index = len(client.log_messages)

    messages = [
        f"{textwrap.indent(m.message, levels[m.type.value - 1])}"
        for m in captured_messages
    ]

    if len(messages) > 0:
        item.add_report_section(call.when, "window/logMessages", "\n".join(messages))


# anext() was added in 3.10
if sys.version_info < (3, 10):

    async def anext(it):  # noqa: A001
        return await it.__anext__()


def get_fixture_arguments(
    fn: Callable,
    client: JsonRPCClient,
    request: pytest.FixtureRequest,
) -> dict:
    """Return the arguments to pass to the user's fixture function.

    Parameters
    ----------
    fn
       The user's fixture function

    client
       The test client instance to inject

    request
       pytest's request fixture

    Returns
    -------
    dict
       The set of arguments to pass to the user's fixture function
    """
    kwargs: dict[str, Any] = {}
    required_parameters = set(inspect.signature(fn).parameters.keys())

    # Inject the 'request' fixture if requested
    if "request" in required_parameters:
        kwargs["request"] = request
        required_parameters.remove("request")

    # Inject the language client
    for name, cls in typing.get_type_hints(fn).items():
        if issubclass(cls, JsonRPCClient):
            kwargs[name] = client
            required_parameters.remove(name)

    # Assume all remaining parameters are pytest fixtures
    for name in required_parameters:
        kwargs[name] = request.getfixturevalue(name)

    return kwargs


def fixture(
    fixture_function=None,
    *,
    config: ClientServerConfig,
    **kwargs,
):
    """Define a fixture that returns a client connected to a server running in a
    background sub-process

    Parameters
    ----------
    config
       Configuration for the client and server.
    """

    def wrapper(fn):
        # pytest-asyncio says `loop_scope` should be present and match `scope`
        # https://pytest-asyncio.readthedocs.io/en/latest/how-to-guides/migrate_from_0_23.html
        if "scope" in kwargs:
            kwargs["loop_scope"] = kwargs["scope"]

        @pytest_asyncio.fixture(**kwargs)
        async def the_fixture(request):
            devtools = request.config.getoption("devtools")
            client = await config.start(devtools=devtools)

            kwargs = get_fixture_arguments(fn, client, request)
            result = fn(**kwargs)
            if inspect.isasyncgen(result):
                try:
                    await anext(result)
                except StopAsyncIteration:
                    pass

            yield client

            if inspect.isasyncgen(result):
                try:
                    await anext(result)
                except StopAsyncIteration:
                    pass

            await client.stop()

        return the_fixture

    if fixture_function:
        return wrapper(fixture_function)

    return wrapper



from lsprotocol.types import TEXT_DOCUMENT_COMPLETION, CompletionItem, CompletionParams
from pygls.lsp.server import LanguageServer

server = LanguageServer("hello-world", "v1")


@server.feature(TEXT_DOCUMENT_COMPLETION)
def completion(ls: LanguageServer, params: CompletionParams):
    return [
        CompletionItem(label="hello"),
        CompletionItem(label="world"),
    ]



from lsprotocol import types
from pygls.lsp.server import LanguageServer

server = LanguageServer("workspace-configuration", "v1")


@server.command("server.configuration")
async def configuration(ls: LanguageServer, *args):
    results = await ls.workspace_configuration_async(
        types.ConfigurationParams(
            items=[
                types.ConfigurationItem(scope_uri="file://workspace/file.txt"),
                types.ConfigurationItem(section="not.found"),
                types.ConfigurationItem(section="values.c"),
            ]
        )
    )

    a = results[0]["values"]["a"]
    assert results[1] is None
    c = results[2]

    return a + c


if __name__ == "__main__":
    server.start_io()



from lsprotocol.converters import get_converter
from pygls.lsp.server import LanguageServer

converter = get_converter()
server = LanguageServer(name="capabilities-server", version="v1.0")


@server.command("return.client.capabilities")
def on_initialize(ls: LanguageServer, *args):
    return ls.client_capabilities


if __name__ == "__main__":
    server.start_io()



from typing import Any
from typing import Optional

import pytest
from lsprotocol import types

from pytest_lsp import checks

a_range = types.Range(
    start=types.Position(line=1, character=0),
    end=types.Position(line=2, character=0),
)


@pytest.mark.parametrize(
    "capabilities,method,params,expected",
    [
        (
            types.ClientCapabilities(
                window=types.WindowClientCapabilities(work_done_progress=False)
            ),
            types.WINDOW_WORK_DONE_PROGRESS_CREATE,
            types.WorkDoneProgressCreateParams(token="id-123"),
            "does not support 'window/workDoneProgress/create'",
        ),
        (
            types.ClientCapabilities(
                window=types.WindowClientCapabilities(work_done_progress=True)
            ),
            types.WINDOW_WORK_DONE_PROGRESS_CREATE,
            types.WorkDoneProgressCreateParams(token="id-123"),
            None,
        ),
        (
            types.ClientCapabilities(
                workspace=types.WorkspaceClientCapabilities(configuration=False)
            ),
            types.WORKSPACE_CONFIGURATION,
            types.ConfigurationParams(items=[]),
            "does not support 'workspace/configuration'",
        ),
        (
            types.ClientCapabilities(
                workspace=types.WorkspaceClientCapabilities(configuration=True)
            ),
            types.WORKSPACE_CONFIGURATION,
            types.ConfigurationParams(items=[]),
            None,
        ),
    ],
)
def test_params_check_warning(
    capabilities: types.ClientCapabilities,
    method: str,
    params: Any,
    expected: Optional[str],
    recwarn,
):
    """Ensure that parameter checks work as expected.

    Parameters
    ----------
    capabilities
       The client's capabilities

    method
       The method name to check

    params
       The params to check

    expected
       The expected warning message

    recwarn
       Builtin fixture from pytest for recording warnings
    """

    if expected is None:
        checks.check_params_against_client_capabilities(capabilities, method, params)
        assert len(recwarn) == 0

    else:
        with pytest.warns(checks.LspSpecificationWarning, match=expected):
            checks.check_params_against_client_capabilities(
                capabilities, method, params
            )


@pytest.mark.parametrize(
    "capabilities,method,result,expected",
    [
        (
            types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    completion=types.CompletionClientCapabilities()
                )
            ),
            types.TEXT_DOCUMENT_COMPLETION,
            [types.CompletionItem(label="item", commit_characters=["."])],
            "does not support commit characters",
        ),
        (
            types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    completion=types.CompletionClientCapabilities(
                        completion_item=types.ClientCompletionItemOptions(
                            commit_characters_support=True
                        )
                    )
                )
            ),
            types.TEXT_DOCUMENT_COMPLETION,
            [types.CompletionItem(label="item", commit_characters=["."])],
            None,
        ),
        (
            types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    completion=types.CompletionClientCapabilities()
                )
            ),
            types.TEXT_DOCUMENT_COMPLETION,
            [
                types.CompletionItem(
                    label="item",
                    documentation=types.MarkupContent(
                        value="", kind=types.MarkupKind.Markdown
                    ),
                )
            ],
            "does not support documentation format 'markdown'",
        ),
        (
            types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    completion=types.CompletionClientCapabilities(
                        completion_item=types.ClientCompletionItemOptions(
                            documentation_format=[types.MarkupKind.Markdown]
                        )
                    )
                )
            ),
            types.TEXT_DOCUMENT_COMPLETION,
            [
                types.CompletionItem(
                    label="item",
                    documentation=types.MarkupContent(
                        value="", kind=types.MarkupKind.Markdown
                    ),
                )
            ],
            None,
        ),
        (
            types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    completion=types.CompletionClientCapabilities()
                )
            ),
            types.TEXT_DOCUMENT_COMPLETION,
            [
                types.CompletionItem(
                    label="item",
                    insert_text_format=types.InsertTextFormat.Snippet,
                )
            ],
            "does not support snippets",
        ),
        (
            types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    completion=types.CompletionClientCapabilities(
                        completion_item=types.ClientCompletionItemOptions(
                            snippet_support=True
                        )
                    )
                )
            ),
            types.TEXT_DOCUMENT_COMPLETION,
            [
                types.CompletionItem(
                    label="item",
                    insert_text_format=types.InsertTextFormat.Snippet,
                )
            ],
            None,
        ),
        (
            types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    document_link=types.DocumentLinkClientCapabilities()
                )
            ),
            types.TEXT_DOCUMENT_DOCUMENT_LINK,
            [types.DocumentLink(range=a_range, tooltip="a tooltip")],
            "does not support tooltips",
        ),
        (
            types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    document_link=types.DocumentLinkClientCapabilities(
                        tooltip_support=True
                    )
                )
            ),
            types.TEXT_DOCUMENT_DOCUMENT_LINK,
            [types.DocumentLink(range=a_range, tooltip="a tooltip")],
            None,
        ),
    ],
)
def test_result_check_warning(
    capabilities: types.ClientCapabilities,
    method: str,
    result: Any,
    expected: Optional[str],
    recwarn,
):
    """Ensure that parameter checks work as expected.

    Parameters
    ----------
    capabilities
       The client's capabilities

    method
       The method name to check

    params
       The params to check

    expected
       The expected warning message

    recwarn
       Builtin fixture from pytest for recording warnings
    """

    if expected is None:
        checks.check_result_against_client_capabilities(capabilities, method, result)
        assert len(recwarn) == 0

    else:
        with pytest.warns(checks.LspSpecificationWarning, match=expected):
            checks.check_result_against_client_capabilities(
                capabilities, method, result
            )

        checks.check_result_against_client_capabilities(capabilities, method, result)



from .checks import LspSpecificationWarning
from .client import LanguageClient
from .client import __version__
from .client import client_capabilities
from .client import make_test_lsp_client
from .plugin import ClientServerConfig
from .plugin import fixture
from .plugin import pytest_addoption
from .plugin import pytest_runtest_makereport
from .protocol import LanguageClientProtocol

__all__ = [
    "__version__",
    "ClientServerConfig",
    "LanguageClient",
    "LanguageClientProtocol",
    "LspSpecificationWarning",
    "client_capabilities",
    "fixture",
    "make_test_lsp_client",
    "pytest_addoption",
    "pytest_runtest_makereport",
]



import pathlib

import pytest

EXAMPLES = pathlib.Path(__file__).parent / "examples"


def setup_test(pytester: pytest.Pytester, example_name: str):
    pytester.makeini(
        """\
[pytest]
asyncio_mode = auto
asyncio_default_fixture_loop_scope = function
"""
    )

    for src in (EXAMPLES / example_name).glob("*.py"):
        if not src.is_file():
            continue

        dest = pytester.path / src.name.replace("t_", "test_")
        dest.write_text(src.read_text())


@pytest.mark.parametrize(
    "name, expected",
    [
        pytest.param("diagnostics", dict(passed=1, warnings=0), id="diagnostics"),
        pytest.param(
            "getting-started", dict(passed=1, warnings=0), id="getting-started"
        ),
        pytest.param(
            "fixture-passthrough", dict(passed=1, warnings=0), id="fixture-passthrough"
        ),
        pytest.param("fixture-scope", dict(passed=2, warnings=0), id="fixture-scope"),
        pytest.param(
            "parameterised-clients",
            dict(passed=2, warnings=0),
            id="parameterised-clients",
        ),
        pytest.param(
            "window-log-message", dict(passed=1, warnings=0), id="window-log-message"
        ),
        pytest.param(
            "window-create-progress",
            dict(passed=3, warnings=0),
            id="window-create-progress",
        ),
        pytest.param(
            "window-show-document",
            dict(passed=1, warnings=0),
            id="window-show-document",
        ),
        pytest.param(
            "window-show-message", dict(passed=1, warnings=0), id="window-show-message"
        ),
        pytest.param(
            "workspace-configuration",
            dict(passed=1, warnings=1),
            id="workspace-configuration",
        ),
    ],
)
def test_examples(pytester: pytest.Pytester, name: str, expected: dict):
    """Ensure that the examples included in the documentation work as expected."""

    setup_test(pytester, name)

    results = pytester.runpytest()
    results.assert_outcomes(**expected)


def test_client_capabilities(pytester: pytest.Pytester):
    """Ensure that the client capabilities example warns as expected."""

    setup_test(pytester, "client-capabilities")

    results = pytester.runpytest()
    results.assert_outcomes(passed=1, warnings=1)

    message = "*LspSpecificationWarning: Client does not support snippets."
    results.stdout.fnmatch_lines(message)


def test_client_capabilities_error(pytester: pytest.Pytester):
    """Ensure that the client capabilities warnings can be upgraded to errors."""

    setup_test(pytester, "client-capabilities")

    results = pytester.runpytest("-W", "error::pytest_lsp.LspSpecificationWarning")
    results.assert_outcomes(failed=1)

    message = (
        "E*pytest_lsp.checks.LspSpecificationWarning: Client does not support snippets."
    )
    results.stdout.fnmatch_lines(message)


def test_client_capabilities_ignore(pytester: pytest.Pytester):
    """Ensure that the client capabilities warnings can be ignored."""

    setup_test(pytester, "client-capabilities")

    results = pytester.runpytest(
        "-W",
        "ignore::pytest_lsp.LspSpecificationWarning",
    )
    results.assert_outcomes(passed=1, warnings=0)


def test_getting_started_fail(pytester: pytest.Pytester):
    """Ensure that the initial getting started example fails as expected."""

    setup_test(pytester, "getting-started-fail")

    results = pytester.runpytest()
    results.assert_outcomes(errors=1)

    message = r"E\s+RuntimeError: Server process \d+ exited with code: 0"
    results.stdout.re_match_lines(message)


def test_generic_rpc(pytester: pytest.Pytester):
    """Ensure that the generic rpc example works as expected"""

    setup_test(pytester, "generic-rpc")

    results = pytester.runpytest("--log-cli-level", "info")
    results.assert_outcomes(passed=1, failed=1)

    results.stdout.fnmatch_lines(" *LOG: a=1")
    results.stdout.fnmatch_lines(" *LOG: b=2")


def test_server_stderr_fail(pytester: pytest.Pytester):
    """Ensure that the server's stderr stream is presented on failure."""

    setup_test(pytester, "server-stderr")

    results = pytester.runpytest()
    results.assert_outcomes(failed=1)

    results.stdout.fnmatch_lines("-* Captured stderr call -*")
    results.stdout.fnmatch_lines("Suggesting item 0")
    results.stdout.fnmatch_lines("Suggesting item 1")
    results.stdout.fnmatch_lines("Suggesting item 2")
    results.stdout.fnmatch_lines("Suggesting item 3")
    results.stdout.fnmatch_lines("Suggesting item 4")
    results.stdout.fnmatch_lines("Suggesting item 5")
    results.stdout.fnmatch_lines("Suggesting item 6")
    results.stdout.fnmatch_lines("Suggesting item 7")
    results.stdout.fnmatch_lines("Suggesting item 8")
    results.stdout.fnmatch_lines("Suggesting item 9")


def test_window_log_message_fail(pytester: pytest.Pytester):
    """Ensure that the initial getting started example fails as expected."""

    setup_test(pytester, "window-log-message-fail")

    results = pytester.runpytest()
    results.assert_outcomes(failed=1)

    results.stdout.fnmatch_lines("-* Captured window/logMessages call -*")
    results.stdout.fnmatch_lines(" *LOG: Suggesting item 0")
    results.stdout.fnmatch_lines(" *LOG: Suggesting item 1")
    results.stdout.fnmatch_lines(" *LOG: Suggesting item 2")
    results.stdout.fnmatch_lines(" *LOG: Suggesting item 3")
    results.stdout.fnmatch_lines(" *LOG: Suggesting item 4")
    results.stdout.fnmatch_lines(" *LOG: Suggesting item 5")
    results.stdout.fnmatch_lines(" *LOG: Suggesting item 6")
    results.stdout.fnmatch_lines(" *LOG: Suggesting item 7")
    results.stdout.fnmatch_lines(" *LOG: Suggesting item 8")
    results.stdout.fnmatch_lines(" *LOG: Suggesting item 9")



import sys

import pytest
import pytest_lsp
from lsprotocol import types
from pytest_lsp import ClientServerConfig, LanguageClient, LspSpecificationWarning


@pytest_lsp.fixture(
    config=ClientServerConfig(server_command=[sys.executable, "server.py"]),
)
async def client(lsp_client: LanguageClient):
    # Setup
    params = types.InitializeParams(capabilities=types.ClientCapabilities())
    await lsp_client.initialize_session(params)

    yield

    # Teardown
    await lsp_client.shutdown_session()


@pytest.mark.asyncio
async def test_progress(client: LanguageClient):
    result = await client.workspace_execute_command_async(
        params=types.ExecuteCommandParams(command="do.progress")
    )

    assert result == "a result"

    progress = client.progress_reports["a-token"]
    assert progress == [
        types.WorkDoneProgressBegin(title="Indexing", percentage=0),
        types.WorkDoneProgressReport(message="25%", percentage=25),
        types.WorkDoneProgressReport(message="50%", percentage=50),
        types.WorkDoneProgressReport(message="75%", percentage=75),
        types.WorkDoneProgressEnd(message="Finished"),
    ]


@pytest.mark.asyncio
async def test_duplicate_progress(client: LanguageClient):
    with pytest.warns(
        LspSpecificationWarning, match="Duplicate progress token: 'duplicate-token'"
    ):
        await client.workspace_execute_command_async(
            params=types.ExecuteCommandParams(command="duplicate.progress")
        )


@pytest.mark.asyncio
async def test_unknown_progress(client: LanguageClient):
    with pytest.warns(
        LspSpecificationWarning, match="Unknown progress token: 'undefined-token'"
    ):
        await client.workspace_execute_command_async(
            params=types.ExecuteCommandParams(command="no.progress")
        )



import sys

import pytest
import pytest_lsp
from lsprotocol.types import (
    TEXT_DOCUMENT_PUBLISH_DIAGNOSTICS,
    ClientCapabilities,
    DidOpenTextDocumentParams,
    InitializeParams,
    TextDocumentItem,
)
from pytest_lsp import ClientServerConfig, LanguageClient


@pytest_lsp.fixture(
    config=ClientServerConfig(server_command=[sys.executable, "server.py"]),
)
async def client(lsp_client: LanguageClient):
    # Setup
    params = InitializeParams(capabilities=ClientCapabilities())
    await lsp_client.initialize_session(params)

    yield

    # Teardown
    await lsp_client.shutdown_session()


@pytest.mark.asyncio
async def test_diagnostics(client: LanguageClient):
    """Ensure that the server implements diagnostics correctly."""

    test_uri = "file:///path/to/file.txt"
    client.text_document_did_open(
        DidOpenTextDocumentParams(
            text_document=TextDocumentItem(
                uri=test_uri,
                language_id="plaintext",
                version=1,
                text="The file's contents",
            )
        )
    )

    # Wait for the server to publish its diagnostics
    await client.wait_for_notification(TEXT_DOCUMENT_PUBLISH_DIAGNOSTICS)

    assert test_uri in client.diagnostics
    assert client.diagnostics[test_uri][0].message == "There is an error here."



from lsprotocol.types import (
    TEXT_DOCUMENT_COMPLETION,
    CompletionItem,
    CompletionParams,
    InsertTextFormat,
)
from pygls.lsp.server import LanguageServer

server = LanguageServer("hello-world", "v1")


@server.feature(TEXT_DOCUMENT_COMPLETION)
def completion(ls: LanguageServer, params: CompletionParams):
    return [
        CompletionItem(
            label="greet",
            insert_text='"Hello, ${1:name}!"$0',
            insert_text_format=InsertTextFormat.Snippet,
        ),
    ]


if __name__ == "__main__":
    server.start_io()



import itertools
import json
import pathlib
import sys
from typing import Any
from typing import Dict
from typing import Optional

import pygls.uris as uri
import pytest

import pytest_lsp
from pytest_lsp import LanguageClient


@pytest.mark.parametrize(
    "client_spec,capabilities",
    [
        *itertools.product(
            ["visual_studio_code", "visual-studio-code"],
            ["visual_studio_code_v1.65.2.json"],
        ),
        *itertools.product(
            ["neovim@0.6", "neovim@v0.6", "neovim@0.6.1"],
            ["neovim_v0.6.1.json"],
        ),
        *itertools.product(
            ["neovim", "neovim@latest", "neovim@v0", "neovim@v0.10", "neovim@v0.10.0"],
            ["neovim_v0.10.0.json"],
        ),
    ],
)
def test_client_capabilities(
    pytester: pytest.Pytester, client_spec: str, capabilities: str
):
    """Ensure that the plugin can mimic the requested client's capabilities correctly.

    Parameters
    ----------
    pytester
       pytest's built in pytester fixture.

    client_spec
       The string used to select the desired client and version

    client_capabilities
       The filename containing the expected client capabilities
    """

    python = sys.executable
    testdir = pathlib.Path(__file__).parent
    server = testdir / "servers" / "capabilities.py"
    root_uri = uri.from_fs_path(str(testdir))

    clients_dir = pathlib.Path(pytest_lsp.__file__).parent / "clients"
    with (clients_dir / capabilities).open() as f:
        # Easiest way to reformat the JSON onto a single line
        expected = json.dumps(json.load(f)["capabilities"])

    pytester.makeini(
        """
        [pytest]
        asyncio_mode = auto
        asyncio_default_fixture_loop_scope = function
        """
    )

    pytester.makeconftest(
        f"""
from lsprotocol.types import InitializeParams

import pytest_lsp
from pytest_lsp import ClientServerConfig
from pytest_lsp import LanguageClient
from pytest_lsp import client_capabilities

@pytest_lsp.fixture(
    config=ClientServerConfig(
        server_command=[r"{python}", r"{server}"],
    )
)
async def client(lsp_client: LanguageClient):
    await lsp_client.initialize_session(
        InitializeParams(
            capabilities=client_capabilities("{client_spec}"),
            root_uri=r"{root_uri}"
        )
    )
    yield

    await lsp_client.shutdown_session()
    """
    )

    pytester.makepyfile(
        f"""
import json
import pytest
from lsprotocol import types
from lsprotocol.converters import get_converter

@pytest.mark.asyncio
async def test_capabilities(client):
    actual = await client.workspace_execute_command_async(
        types.ExecuteCommandParams(command="return.client.capabilities")
    )

    expected = json.loads('{expected}')

    # lsprotocol is going to filter out any quirks of the client
    # so we can't compare the dicts directly.
    converter = get_converter()
    actual_capabilities = converter.structure(actual, types.ClientCapabilities)
    expected_capabilities = converter.structure(expected, types.ClientCapabilities)
    assert actual_capabilities == expected_capabilities
    """
    )

    results = pytester.runpytest("-vv")
    results.assert_outcomes(passed=1, warnings=0)


@pytest.mark.parametrize(
    "config,section,scope,expected",
    [
        ({"": 12}, None, None, 12),
        ({"": 12}, "example.section", None, None),
        ({"": 12}, None, "file://path/to/workspace/file.txt", 12),
        ({"": {"example": {"section": 12}}}, "example", None, {"section": 12}),
        ({"": {"example": {"section": 12}}}, "example.section", None, 12),
        (
            {"": {"example": {"section": 12}}},
            "example.section",
            "file://path/to/workspace/file.txt",
            12,
        ),
        (
            {
                "": {"example": {"section": 12}},
                "file://path/to/workspace": {"example": {"section": 32}},
            },
            "example.section",
            "file://path/to/workspace/file.txt",
            32,
        ),
        # To keep things simple, we will not try to fallback to more general scopes.
        (
            {
                "": {"example": {"section": 12}},
                "file://path/to/workspace": {"example": 32},
            },
            "example.section",
            "file://path/to/workspace/file.txt",
            None,
        ),
    ],
)
def test_get_configuration(
    config: Dict[str, Any], section: Optional[str], scope: Optional[str], expected: Any
):
    """Ensure that we can get a client's configuration correctly.

    Parameters
    ----------
    config
       The client's configuration

    section
       The section name to retrieve

    scope
       The scope to retrieve

    expected
       The expected result
    """

    client = LanguageClient(configuration=config)
    assert client.get_configuration(section=section, scope_uri=scope) == expected


@pytest.mark.parametrize(
    "item,section,scope,expected",
    [
        (12, None, None, {"": 12}),
        (
            12,
            None,
            "file://path/to/workspace",
            {"": {}, "file://path/to/workspace": 12},
        ),
        (12, "example", None, {"": {"example": 12}}),
        (
            12,
            "example.config.section",
            None,
            {"": {"example": {"config": {"section": 12}}}},
        ),
    ],
)
def test_set_configuration(
    item: Any, section: Optional[str], scope: Optional[str], expected: Dict[str, Any]
):
    """Ensure that we can set the client's configuration correctly.

    Parameters
    ----------
    item
       The value to set

    section
       The configuration section to set

    scope
       The scope at which to set the configuration

    expected
       The expected result
    """
    client = LanguageClient()
    client.set_configuration(item, section=section, scope_uri=scope)

    assert client._configuration == expected


def test_set_configuration_equivalent_methods():
    """Ensure that setting the configuration using alternate, but equivalent methods
    yields the same result."""

    client_one = LanguageClient()
    client_two = LanguageClient()

    config = {
        "example": {
            "section": 12,
            "value": "test",
            "nested": {"section": 34},
        },
    }
    client_one.set_configuration(config)

    client_two.set_configuration(12, section="example.section")
    client_two.set_configuration("test", section="example.value")
    client_two.set_configuration(34, section="example.nested.section")

    expected = {
        "": {
            "example": {
                "section": 12,
                "value": "test",
                "nested": {"section": 34},
            },
        },
    }
    assert client_one._configuration == expected == client_two._configuration



import sys

import pytest
import pytest_lsp
from lsprotocol import types
from pytest_lsp import ClientServerConfig, LanguageClient


@pytest_lsp.fixture(
    config=ClientServerConfig(server_command=[sys.executable, "server.py"]),
)
async def client(lsp_client: LanguageClient):
    # Setup
    params = types.InitializeParams(
        capabilities=types.ClientCapabilities(
            workspace=types.WorkspaceClientCapabilities(configuration=False)
        )
    )
    await lsp_client.initialize_session(params)

    yield

    # Teardown
    await lsp_client.shutdown_session()


@pytest.mark.asyncio
async def test_configuration(client: LanguageClient):
    global_config = {"values": {"a": 42, "c": 4}}

    workspace_uri = "file://workspace/file.txt"
    workspace_config = {"a": 1, "c": 1}

    client.set_configuration(global_config)
    client.set_configuration(
        workspace_config, section="values", scope_uri=workspace_uri
    )

    result = await client.workspace_execute_command_async(
        params=types.ExecuteCommandParams(command="server.configuration")
    )
    assert result == 5



import sys

import pytest_lsp
from lsprotocol.types import (
    ClientCapabilities,
    CompletionList,
    CompletionParams,
    InitializeParams,
    Position,
    TextDocumentIdentifier,
)
from pytest_lsp import ClientServerConfig, LanguageClient


@pytest_lsp.fixture(
    config=ClientServerConfig(server_command=[sys.executable, "server.py"]),
)
async def client(lsp_client: LanguageClient):
    # Setup
    params = InitializeParams(capabilities=ClientCapabilities())
    await lsp_client.initialize_session(params)

    yield

    # Teardown
    await lsp_client.shutdown_session()


async def test_completions(client: LanguageClient):
    """Ensure that the server implements completions correctly."""

    results = await client.text_document_completion_async(
        params=CompletionParams(
            position=Position(line=1, character=0),
            text_document=TextDocumentIdentifier(uri="file:///path/to/file.txt"),
        )
    )

    assert results is not None

    if isinstance(results, CompletionList):
        items = results.items
    else:
        items = results

    labels = [item.label for item in items]
    assert labels == ["hello", "world"]



"""Checks to ensure that the server is compliant with the spec.

These can be used in your unit tests to ensure that objects (such as ``CompletionItem``
etc) take into account the client's capabilities when constructed.

However, they are also called automatically during end-to-end tests that make use of the
standard :class:`~pytest_lsp.LanguageClient`. See :ref:`pytest-lsp-spec-checks` for more
details.

"""

from __future__ import annotations

# ruff: noqa: S101
import logging
import warnings
from typing import Any
from typing import Callable
from typing import Optional
from typing import Union

from lsprotocol import types
from pygls.capabilities import get_capability

logger = logging.getLogger(__name__)
ParamsChecker = Callable[[types.ClientCapabilities, Any], None]
ResultChecker = Callable[[types.ClientCapabilities, Any], None]

PARAMS_CHECKS: dict[str, ParamsChecker] = {}
RESULT_CHECKS: dict[str, ResultChecker] = {}


class LspSpecificationWarning(UserWarning):
    """Warning raised when encountering results that fall outside the spec."""


def check_result_for(*, method: str) -> Callable[[ResultChecker], ResultChecker]:
    """Define a result check."""

    def defcheck(fn: ResultChecker):
        if (existing := RESULT_CHECKS.get(method)) is not None:
            raise ValueError(f"{fn!r} conflicts with existing check {existing!r}")

        RESULT_CHECKS[method] = fn
        return fn

    return defcheck


def check_params_of(*, method: str) -> Callable[[ParamsChecker], ParamsChecker]:
    """Define a params check."""

    def defcheck(fn: ParamsChecker):
        if (existing := PARAMS_CHECKS.get(method)) is not None:
            raise ValueError(f"{fn!r} conflicts with existing check {existing!r}")

        PARAMS_CHECKS[method] = fn
        return fn

    return defcheck


def check_result_against_client_capabilities(
    capabilities: Optional[types.ClientCapabilities], method: str, result: Any
):
    """Check that the given result respects the client's declared capabilities.

    This will emit an ``LspSpecificationWarning`` if any issues are detected.

    Parameters
    ----------
    capabilities
       The client's capabilities

    method
       The method name to validate the result of

    result
       The result to validate
    """

    if capabilities is None:
        raise RuntimeError("Client has not been initialized")

    # Only run checks if the user provided some capabilities for the client.
    if capabilities == types.ClientCapabilities():
        return

    result_checker = RESULT_CHECKS.get(method)
    if result_checker is None:
        return

    try:
        result_checker(capabilities, result)
    except AssertionError as e:
        warnings.warn(str(e), LspSpecificationWarning, stacklevel=4)


def check_params_against_client_capabilities(
    capabilities: Optional[types.ClientCapabilities], method: str, params: Any
):
    """Check that the given params respect the client's declared capabilities.

    This will emit an ``LspSpecificationWarning`` if any issues are detected.

    Parameters
    ----------
    capabilities
       The client's capabilities

    method
       The method name to validate the result of

    params
       The params to validate
    """
    if capabilities is None:
        raise RuntimeError("Client has not been initialized")

    # Only run checks if the user provided some capabilities for the client.
    if capabilities == types.ClientCapabilities():
        return

    params_checker = PARAMS_CHECKS.get(method)
    if params_checker is None:
        return

    try:
        params_checker(capabilities, params)
    except AssertionError as e:
        warnings.warn(str(e), LspSpecificationWarning, stacklevel=2)


def check_completion_item(
    item: types.CompletionItem,
    commit_characters_support: bool,
    documentation_formats: set[str],
    snippet_support: bool,
):
    """Ensure that the given ``CompletionItem`` complies with the given capabilities."""

    if item.commit_characters:
        assert commit_characters_support, "Client does not support commit characters"

    if isinstance(item.documentation, types.MarkupContent):
        kind = item.documentation.kind
        message = f"Client does not support documentation format {kind.value!r}"
        assert kind in documentation_formats, message

    if item.insert_text_format == types.InsertTextFormat.Snippet:
        assert snippet_support, "Client does not support snippets."


@check_result_for(method=types.TEXT_DOCUMENT_COMPLETION)
def completion_items(
    capabilities: types.ClientCapabilities,
    result: Union[types.CompletionList, list[types.CompletionItem], None],
):
    """Ensure that the completion items returned from the server are compliant with the
    spec and the client's declared capabilities."""

    if result is None:
        return

    commit_characters_support = get_capability(
        capabilities,
        "text_document.completion.completion_item.commit_characters_support",
        False,
    )
    documentation_formats = set(
        get_capability(
            capabilities,
            "text_document.completion.completion_item.documentation_format",
            [],
        )
    )
    snippet_support = get_capability(
        capabilities,
        "text_document.completion.completion_item.snippet_support",
        False,
    )

    if isinstance(result, types.CompletionList):
        items = result.items
    else:
        items = result

    for item in items:
        check_completion_item(
            item,
            commit_characters_support=commit_characters_support,
            documentation_formats=documentation_formats,
            snippet_support=snippet_support,
        )


@check_result_for(method=types.COMPLETION_ITEM_RESOLVE)
def completion_item_resolve(
    capabilities: types.ClientCapabilities, item: types.CompletionItem
):
    """Ensure that the completion item returned from the server is compliant with the
    spec and the client's declared capbabilities."""

    commit_characters_support = get_capability(
        capabilities,
        "text_document.completion.completion_item.commit_characters_support",
        False,
    )
    documentation_formats = set(
        get_capability(
            capabilities,
            "text_document.completion.completion_item.documentation_format",
            [],
        )
    )
    snippet_support = get_capability(
        capabilities,
        "text_document.completion.completion_item.snippet_support",
        False,
    )

    check_completion_item(
        item,
        commit_characters_support=commit_characters_support,
        documentation_formats=documentation_formats,
        snippet_support=snippet_support,
    )


@check_result_for(method=types.TEXT_DOCUMENT_DOCUMENT_LINK)
def document_links(
    capabilities: types.ClientCapabilities, result: Optional[list[types.DocumentLink]]
):
    """Ensure that the document links returned from the server are compliant with the
    Spec and the client's declared capabilities."""

    if result is None:
        return

    tooltip_support = get_capability(
        capabilities, "text_document.document_link.tooltip_support", False
    )

    for item in result:
        if item.tooltip:
            assert tooltip_support, "Client does not support tooltips."


@check_params_of(method=types.WINDOW_WORK_DONE_PROGRESS_CREATE)
def work_done_progress_create(
    capabilities: types.ClientCapabilities,
    params: types.WorkDoneProgressCreateParams,
):
    """Assert that the client has support for ``window/workDoneProgress/create``
    requests."""
    is_supported = get_capability(capabilities, "window.work_done_progress", False)
    assert is_supported, "Client does not support 'window/workDoneProgress/create'"


@check_params_of(method=types.WORKSPACE_CONFIGURATION)
def workspace_configuration(
    capabilities: types.ClientCapabilities,
    params: types.ConfigurationParams,
):
    """Ensure that the client has support for ``workspace/configuration`` requests."""
    is_supported = get_capability(capabilities, "workspace.configuration", False)
    assert is_supported, "Client does not support 'workspace/configuration'"



from pygls.protocol import JsonRPCProtocol, default_converter
from pygls.server import JsonRPCServer

server = JsonRPCServer(
    protocol_cls=JsonRPCProtocol, converter_factory=default_converter
)


@server.feature("math/add")
def addition(ls: JsonRPCServer, params):
    a = params.a
    b = params.b

    ls.protocol.notify("log/message", dict(message=f"{a=}"))
    ls.protocol.notify("log/message", dict(message=f"{b=}"))

    return dict(total=a + b)


@server.feature("math/sub")
def subtraction(ls: JsonRPCServer, params):
    a = params.a
    b = params.b

    ls.protocol.notify("log/message", dict(message=f"{a=}"))
    ls.protocol.notify("log/message", dict(message=f"{b=}"))

    return dict(total=b - a)


if __name__ == "__main__":
    server.start_io()



from __future__ import annotations

import asyncio
import json
import logging
import os
import pathlib
import sys
import traceback
import typing
import warnings
from importlib import resources  # type: ignore[no-redef]

from lsprotocol import types
from lsprotocol.converters import get_converter
from packaging.version import parse as parse_version
from pygls.exceptions import JsonRpcException
from pygls.exceptions import PyglsError
from pygls.lsp.client import BaseLanguageClient
from pygls.protocol import default_converter

from .checks import LspSpecificationWarning
from .protocol import LanguageClientProtocol

if typing.TYPE_CHECKING:
    from typing import Any


__version__ = "1.0.0b1"
logger = logging.getLogger(__name__)


class LanguageClient(BaseLanguageClient):
    """Used to drive language servers under test."""

    protocol: LanguageClientProtocol

    def __init__(self, *args, configuration: dict[str, Any] | None = None, **kwargs):
        if "protocol_cls" not in kwargs:
            kwargs["protocol_cls"] = LanguageClientProtocol

        super().__init__("pytest-lsp-client", __version__, *args, **kwargs)

        self.capabilities: types.ClientCapabilities | None = None
        """The client's capabilities."""

        self.shown_documents: list[types.ShowDocumentParams] = []
        """Holds any received show document requests."""

        self.messages: list[types.ShowMessageParams] = []
        """Holds any received ``window/showMessage`` requests."""

        self.log_messages: list[types.LogMessageParams] = []
        """Holds any received ``window/logMessage`` requests."""

        self.diagnostics: dict[str, list[types.Diagnostic]] = {}
        """Holds any recieved diagnostics."""

        self.progress_reports: dict[
            types.ProgressToken, list[types.ProgressParams]
        ] = {}
        """Holds any received progress updates."""

        self.error: Exception | None = None
        """Indicates if the client encountered an error."""

        config = (configuration or {"": {}}).copy()
        if "" not in config:
            config[""] = {}

        self._configuration: dict[str, dict[str, Any]] = config
        """Holds ``workspace/configuration`` values."""

        self._setup_log_index = 0
        """Used to keep track of which log messages occurred during startup."""

        self._last_log_index = 0
        """Used to keep track of which log messages correspond with which test case."""

        self._stderr_forwarder: asyncio.Task | None = None
        """A task that forwards the server's stderr to the test process."""

    async def start_io(self, cmd: str, *args, **kwargs):
        await super().start_io(cmd, *args, **kwargs)

        # Forward the server's stderr to this process' stderr
        if self._server and self._server.stderr:
            self._stderr_forwarder = asyncio.create_task(forward_stderr(self._server))

    async def stop(self):
        if self._stderr_forwarder:
            self._stderr_forwarder.cancel()

        return await super().stop()

    async def server_exit(self, server: asyncio.subprocess.Process):
        """Called when the server process exits."""

        if self._stop_event.is_set():
            return

        # TODO: Should the upstream base client be doing this?
        # Cancel any pending futures.
        reason = f"Server process {server.pid} exited with code: {server.returncode}"

        for id_, fut in self.protocol._request_futures.items():
            if not fut.done():
                fut.set_exception(RuntimeError(reason))
                logger.debug("Cancelled pending request '%s': %s", id_, reason)

    def report_server_error(
        self, error: Exception, source: PyglsError | JsonRpcException
    ):
        """Called when the server does something unexpected, e.g. sending malformed
        JSON."""
        self.error = error
        tb = "".join(traceback.format_exc())

        message = f"{source.__name__}: {error}\n{tb}"  # type: ignore

        loop = asyncio.get_running_loop()
        loop.call_soon(cancel_all_tasks, message)

        if self._stop_event:
            self._stop_event.set()

    def get_configuration(
        self, *, section: str | None = None, scope_uri: str | None = None
    ) -> Any | None:
        """Get a configuration value.

        Parameters
        ----------
        section
           The optional section name to retrieve.
           If ``None`` the top level configuration object for the requested scope will
           be returned

        scope_uri
           The scope at which to set the configuration.
           If ``None``, this will default to the global scope.

        Returns
        -------
        Optional[Any]
           The requested configuration value or ``None`` if not found.
        """
        section = section or ""
        scope = scope_uri or ""

        # Find the longest prefix of ``scope``. The empty string is a prefix of all
        # strings so there will always be at least one match
        candidates = [c for c in self._configuration.keys() if scope.startswith(c)]
        selected = sorted(candidates, key=len, reverse=True)[0]

        if (item := self._configuration.get(selected, None)) is None:
            return None

        if section == "":
            return item

        for segment in section.split("."):
            if not hasattr(item, "get"):
                return None

            if (item := item.get(segment, None)) is None:
                return None

        return item

    def set_configuration(
        self,
        item: Any,
        *,
        section: str | None = None,
        scope_uri: str | None = None,
    ):
        """Set a configuration value.

        Parameters
        ----------
        item
           The value to set

        section
           The optional section name to set.
           If ``None`` the top level configuration object will be overriden with
           ``item``.

        scope_uri
           The scope at which to set the configuration.
           If ``None``, this will default to the global scope.
        """
        section = section or ""
        scope = scope_uri or ""

        if section == "":
            self._configuration[scope] = item
            return

        config = self._configuration.setdefault(scope, {})
        *parents, name = section.split(".")

        for segment in parents:
            config = config.setdefault(segment, {})

        config[name] = item

    async def initialize_session(
        self, params: types.InitializeParams
    ) -> types.InitializeResult:
        """Make an ``initialize`` request to a lanaguage server.

        It will also automatically send an ``initialized`` notification once
        the server responds.

        Parameters
        ----------
        params
           The parameters to send to the client.

           The following fields will be automatically set if left blank.

           - ``process_id``: Set to the PID of the current process.

        Returns
        -------
        InitializeResult
           The result received from the client.
        """
        self.capabilities = params.capabilities

        if params.process_id is None:
            params.process_id = os.getpid()

        response = await self.initialize_async(params)
        self.initialized(types.InitializedParams())

        return response

    async def shutdown_session(self) -> None:
        """Shutdown the server under test.

        Helper method that handles sending ``shutdown`` and ``exit`` messages in the
        correct order.

        .. note::

           This method will not attempt to send these messages if a fatal error has
           occurred.

        """
        if self.error is not None or self.capabilities is None:
            return

        await self.shutdown_async(None)

        self.exit(None)
        if self._server:
            await self._server.wait()

    async def wait_for_notification(self, method: str):
        """Block until a notification with the given method is received.

        Parameters
        ----------
        method
           The notification method to wait for, e.g. ``textDocument/publishDiagnostics``
        """
        return await self.protocol.wait_for_notification_async(method)


async def forward_stderr(server: asyncio.subprocess.Process):
    if server.stderr is None:
        return

    # EOF is signalled with an empty bytestring
    while (line := await server.stderr.readline()) != b"":
        sys.stderr.buffer.write(line)


def cancel_all_tasks(message: str):
    """Called to cancel all awaited tasks."""

    for task in asyncio.all_tasks():
        task.cancel(message)


def make_test_lsp_client() -> LanguageClient:
    """Construct a new test client instance with the handlers needed to capture
    additional responses from the server."""

    client = LanguageClient(
        converter_factory=default_converter,
    )

    @client.feature(types.WORKSPACE_CONFIGURATION)
    def configuration(client: LanguageClient, params: types.ConfigurationParams):
        return [
            client.get_configuration(section=item.section, scope_uri=item.scope_uri)
            for item in params.items
        ]

    @client.feature(types.TEXT_DOCUMENT_PUBLISH_DIAGNOSTICS)
    def publish_diagnostics(
        client: LanguageClient, params: types.PublishDiagnosticsParams
    ):
        client.diagnostics[params.uri] = params.diagnostics

    @client.feature(types.WINDOW_WORK_DONE_PROGRESS_CREATE)
    def create_work_done_progress(
        client: LanguageClient, params: types.WorkDoneProgressCreateParams
    ):
        if params.token in client.progress_reports:
            # TODO: Send an error reponse to the client - might require changes
            #       to pygls...
            warnings.warn(
                f"Duplicate progress token: {params.token!r}",
                LspSpecificationWarning,
                stacklevel=2,
            )

        client.progress_reports.setdefault(params.token, [])

    @client.feature(types.PROGRESS)
    def progress(client: LanguageClient, params: types.ProgressParams):
        if params.token not in client.progress_reports:
            warnings.warn(
                f"Unknown progress token: {params.token!r}",
                LspSpecificationWarning,
                stacklevel=2,
            )

        if not params.value:
            return

        if (kind := params.value.get("kind", None)) == "begin":
            type_: type[Any] = types.WorkDoneProgressBegin
        elif kind == "report":
            type_ = types.WorkDoneProgressReport
        elif kind == "end":
            type_ = types.WorkDoneProgressEnd
        else:
            raise TypeError(f"Unknown progress kind: {kind!r}")

        value = client.protocol._converter.structure(params.value, type_)
        client.progress_reports.setdefault(params.token, []).append(value)

    @client.feature(types.WINDOW_LOG_MESSAGE)
    def log_message(client: LanguageClient, params: types.LogMessageParams):
        client.log_messages.append(params)

        levels = [logger.error, logger.warning, logger.info, logger.debug]
        levels[params.type.value - 1](params.message)

    @client.feature(types.WINDOW_SHOW_MESSAGE)
    def show_message(client: LanguageClient, params):
        client.messages.append(params)

    @client.feature(types.WINDOW_SHOW_DOCUMENT)
    def show_document(
        client: LanguageClient, params: types.ShowDocumentParams
    ) -> types.ShowDocumentResult:
        client.shown_documents.append(params)
        return types.ShowDocumentResult(success=True)

    return client


def client_capabilities(client_spec: str) -> types.ClientCapabilities:
    """Find the capabilities that correspond to the given client spec.

    This function supports the following syntax

    ``client-name`` or ``client-name@latest``
       Return the capabilities of the latest version of ``client-name``

    ``client-name@v2``
       Return the latest release of the ``v2`` of ``client-name``

    ``client-name@v2.3.1``
       Return exactly ``v2.3.1`` of ``client-name``

    Parameters
    ----------
    client_spec
       The string describing the client to load the corresponding
       capabilities for.

    Raises
    ------
    ValueError
       If the requested client's capabilities could not be found

    Returns
    -------
    ClientCapabilities
       The requested client capabilities
    """

    candidates: dict[str, pathlib.Path] = {}

    client_spec = client_spec.replace("-", "_")
    target_version = "latest"

    if "@" in client_spec:
        client_spec, target_version = client_spec.split("@")
        if target_version.startswith("v"):
            target_version = target_version[1:]

    for resource in resources.files("pytest_lsp.clients").iterdir():
        filename = typing.cast(pathlib.Path, resource)

        # Skip the README or any other files that we don't care about.
        if filename.suffix != ".json":
            continue

        name, version = filename.stem.split("_v")
        if name == client_spec:
            if version.startswith(target_version) or target_version == "latest":
                candidates[version] = filename

    if len(candidates) == 0:
        raise ValueError(
            f"Could not find capabilities for '{client_spec}@{target_version}'"
        )

    # Out of the available candidates, choose the latest version
    selected_version = sorted(candidates.keys(), key=parse_version, reverse=True)[0]
    filename = candidates[selected_version]

    converter = get_converter()
    capabilities = json.loads(filename.read_text())

    params = converter.structure(capabilities, types.InitializeParams)
    logger.info(
        "Selected %s v%s",
        params.client_info.name,  # type: ignore[union-attr]
        params.client_info.version,  # type: ignore[union-attr]
    )

    return params.capabilities



# A server that exits mid request.
import sys

from lsprotocol import types
from pygls.lsp.server import LanguageServer


class CountingLanguageServer(LanguageServer):
    count: int = 0


server = CountingLanguageServer(name="completion-exit-server", version="v1.0")


@server.feature(types.TEXT_DOCUMENT_COMPLETION)
def on_complete(server: CountingLanguageServer, params: types.CompletionParams):
    server.count += 1
    if server.count == 5:
        sys.exit(0)

    return [types.CompletionItem(label=f"{server.count}")]


if __name__ == "__main__":
    server.start_io()



import logging
import sys

import pytest
import pytest_lsp
from pygls.client import JsonRPCClient
from pytest_lsp import ClientServerConfig


def client_factory():
    client = JsonRPCClient()

    @client.feature("log/message")
    def _on_message(params):
        logging.info("LOG: %s", params.message)

    return client


@pytest_lsp.fixture(
    config=ClientServerConfig(
        client_factory=client_factory, server_command=[sys.executable, "server.py"]
    ),
)
async def client(rpc_client: JsonRPCClient):
    # Setup code here (if any)

    yield

    # Teardown code here (if any)


@pytest.mark.asyncio
async def test_add(client: JsonRPCClient):
    """Ensure that the server implements addition correctly."""

    result = await client.protocol.send_request_async(
        "math/add", params={"a": 1, "b": 2}
    )
    assert result.total == 3


@pytest.mark.asyncio
async def test_sub(client: JsonRPCClient):
    """Ensure that the server implements addition correctly."""

    result = await client.protocol.send_request_async(
        "math/sub", params={"a": 1, "b": 2}
    )
    assert result.total == -1



from lsprotocol import types
from pygls.lsp.server import LanguageServer

server = LanguageServer("diagnostic-server", "v1")


@server.feature(types.TEXT_DOCUMENT_DID_OPEN)
def did_open(ls: LanguageServer, params: types.DidOpenTextDocumentParams):
    ls.text_document_publish_diagnostics(
        types.PublishDiagnosticsParams(
            uri=params.text_document.uri,
            diagnostics=[
                types.Diagnostic(
                    message="There is an error here.",
                    range=types.Range(
                        start=types.Position(line=1, character=1),
                        end=types.Position(line=1, character=10),
                    ),
                )
            ],
        )
    )


if __name__ == "__main__":
    server.start_io()
